/**
	Headers containing the various definitions for the functional partitioning
	version with count based windows
*/
#ifndef _FUNC_PART_H
#define _FUNC_PART_H
#include <pthread.h>
#include "general.h"
#include <ff/buffer.hpp>
#include <ff/allocator.hpp>

/**
	Data structure passed to the various entities
*/
//Struttura dati dello stato dell'emettitore:
typedef struct emitter_data {
	int port; //port from which receive the connection from the generator
	int num_workers; //number of workers
	int num_classes; //number of classes

	ff::SWSR_Ptr_Buffer **outqueue; //queues towards workers
	pthread_barrier_t *barrier; //initial synchronization barrier
	//class_descr_t *schedulingTable; //Tabella delle informazioni per la schedulazione "smart".
	ticks monitoring_step; //time interval between two monitoring phases
	//char *scheduling_table; //scheduling table
	ff::ff_allocator *ffalloc; //fastflow memory allocator
	ticks *start_global_ticks; //start time, derived from the first tuple
	long int freq;
	char *suffix;

	//output queue towards the controller
	ff::SWSR_Ptr_Buffer *cn_outqueue;
	//input queue from the controller
	ff::SWSR_Ptr_Buffer *cn_inqueue;

	//repository for configuration
	void *repo;
} emitter_data_t;

//Struttura dati dello stato del collettore:
typedef struct collector_data {
	int num_workers; //number of workers
	int num_classes; //number of classes
	ff::SWSR_Ptr_Buffer **inqueue; //input queues from workers
	ff::SWSR_Ptr_Buffer **wqueue; //queues emitter-> workers (only for monitoring purposes, read only)
	pthread_barrier_t *barrier; //initial synchronization barrier
	ticks *start_global_ticks; //start time, derived from the first tuple
	//for testing only
	long int freq; //processor frequency
	char *suffix;
	int window_slide;
	//output queue towards the controller
	ff::SWSR_Ptr_Buffer *cn_outqueue;
	ff::SWSR_Ptr_Buffer *cn_inqueue;
	int max_workers;
} collector_data_t;

//Struttura dati dello stato del generico worker:
typedef struct worker_data {
	ff::SWSR_Ptr_Buffer *inqueue; //queue from Emitter
	ff::SWSR_Ptr_Buffer *outqueue; //queue toward Collector
	int workerId; //Id
	pthread_barrier_t *barrier; ////initial synchronization barrier
	int window_size;
	int window_slide;
	ticks *start_global_ticks; //start time, derived from the first tuple
	//double *comp_time; //computation times for the various classes
	ff::ff_allocator *ffalloc; //fastflow memory allocator
	int num_classes;
	//for testing only
	long int freq; //processor frequency

	//output queue towards the controller
	ff::SWSR_Ptr_Buffer *cn_outqueue;
	//repository for configuration
	void *repo;

} worker_data_t;

typedef struct controller_data {
	int num_workers; //number of workers
	int num_classes; //number of classes
	long int freq;
	char *suffix;
	//input queues
	ff::SWSR_Ptr_Buffer *e_inqueue;
	ff::SWSR_Ptr_Buffer **w_inqueue;
	ff::SWSR_Ptr_Buffer *c_inqueue;
	
	//output queues
	ff::SWSR_Ptr_Buffer *e_outqueue;
	ff::SWSR_Ptr_Buffer *c_outqueue;
	ticks *start_global_ticks; //start time, derived from the first tuple
	//repository for configuration
	void *repo;
	int *affinities;
	int max_workers;
	//these will be used for thread spawning
	int window_size;
	int window_slide;
	//double *comp_time; //computation times for the various classes
	ff::ff_allocator *ffalloc; //fastflow memory allocator
	
}controller_data_t;

/**
	Data structure passed by the various entities to the controller.
	They will contain the monitoring data referring to the last monitoring 
	step

	NOTES:
		- times are taken in cycle: for this reason frequency must be kept at maximum
*/

typedef enum monitoring_tag{
	MON_TAG=0,
	RECONF_FINISHED=1
}monitoring_tag;

typedef struct{
	union{	
		struct
		{
			char mon_id; //just for testing
			monitoring_tag tag;
			int64_t elements; //number of elements received in the last monitoring step
			int buffer_elements; //number of elements that are in the tcp buffer while sending monitoring data
			double grow_rate_per_step;
			int* elements_per_class; //numb of elements per class
			bool stop; //used for stopping monitoring
			char *scheduling_table; //actual scheduling table (PUNTATORE O COPIA?)
		};
		char padding[CACHE_LINE_SIZE];
	};
}emitter_monitoring_t;

typedef struct{
	union{	
		struct
		{
			int *elements_per_class; //numb of elements per class (it is redundant but we have to keep it for the worker)
			int *computations_per_class; //numb of computation per class (it is redundant but we have to keep it for the worker and will be helpful in computing the right metrics when the module is bottleneck)
			double *tcalc_per_class; //tcalc expressed in usec (total tcalc per class, not on average)
			int elements_rcvd; //total number of element received
			int computations; //number of computation performed
		};
		char padding[CACHE_LINE_SIZE];
	};
}worker_monitoring_t;

typedef struct{
	union{	
		struct
		{
			char mon_id; //just for testing
			monitoring_tag tag;
			int results; //total received results 
			int *results_per_class; //results received for each class
			double *latency_per_class; //expressed in usec (averaged for each class)
			ticks monitoring_time;
		};
		char padding[CACHE_LINE_SIZE];
	};
}collector_monitoring_t;


/**
	Definition of return types and grains (i.e. internal iteration) for G and H functions. Since their implementation is strictly related to
	the data structure used to keep the elements in window, their definition is in the various data structure header files
	(such as cbwindow.hpp).
*/
typedef double GReturn_type;
typedef struct{
	double res;
	long long timestamp; //timestamp for monitoring purposes 
	int64_t id; //id of the task that has triggered the computation
	bool isEOS; //true if it will represent to the collector the end of the stream
	int type; //class
	char wid; //id of the worker that performed the computation
	ticks ts;
	void *res_buff=NULL; //this will be valid only if it is the EOS sent from a Worker to the Collector (for freeing the result_buffer)
} HReturn_type;


/**
	Function Declarations
*/
void *emitter(void *args);
void * worker(void *args);
void * collector(void *args);
void *controller (void *args);

#endif