/**
	Headers containing the various definitions for the functional partitioning
	version with count based windows
*/
#ifndef _PANE_CB_H
#define _PANE_CB_H
#include <pthread.h>
#include <ff/buffer.hpp>
//Struttura dati dello stato dell'emettitore:
typedef struct emitter_data {
	int port; //port from which receive the connection from the generator
	int num_workers; //number of workers
	int num_classes; //number of classes
	int pane_size;
	ff::SWSR_Ptr_Buffer **outqueue; //queues towards workers
	pthread_barrier_t *barrier; //initial synchronization barrier
	//class_descr_t *schedulingTable; //Tabella delle informazioni per la schedulazione "smart".
	ticks monitoring_step; //time interval between two monitoring phases
	
	
	ff::ff_allocator *ffalloc; //fastflow memory allocator
	ticks *start_global_ticks; //start time, derived from the first tuple
	long int freq;
	char *suffix;

	//output queue towards the controller
	ff::SWSR_Ptr_Buffer *cn_outqueue;
	//input queue from the controller
	ff::SWSR_Ptr_Buffer *cn_inqueue;
} emitter_data_t;

//Struttura dati dello stato del collettore:
typedef struct collector_data {
	int num_workers; //number of workers
	int num_classes; //number of classes
	ff::SWSR_Ptr_Buffer **inqueue; //input queues from workers
	ff::SWSR_Ptr_Buffer **wqueue; //queues emitter-> workers (only for monitoring purposes, read only)
	pthread_barrier_t *barrier; //initial synchronization barrier
	int pane_size;
	int window_size;
	int window_slide;
	ticks *start_global_ticks; //start time, derived from the first tuple
	//for testing only
	long int freq; //processor frequency
	char *suffix;
	//output queue towards the controller
	ff::SWSR_Ptr_Buffer *cn_outqueue;
	//input queue from the controller
	ff::SWSR_Ptr_Buffer *cn_inqueue;

	//double stepLength; //Lunghezza di uno step (secondi).
} collector_data_t;

//Struttura dati dello stato del generico worker:
typedef struct worker_data {
	ff::SWSR_Ptr_Buffer *inqueue; //queue from Emitter
	ff::SWSR_Ptr_Buffer *outqueue; //queue toward Collector
	int workerId; //Id
	pthread_barrier_t *barrier; ////initial synchronization barrier
	int pane_size;
	//double execTime; //Tempo di esecuzione di un task (millisecondi).
	ff::ff_allocator *ffalloc; //fastflow memory allocator
	int num_classes;
	//for testing only
	long int freq; //processor frequency
	//output queue towards the controller
	ff::SWSR_Ptr_Buffer *cn_outqueue;
} worker_data_t;

typedef struct controller_data {
	int num_workers; //number of workers
	int num_classes; //number of classes
	long int freq;
	char *suffix;
	//input queues
	ff::SWSR_Ptr_Buffer *e_inqueue;
	ff::SWSR_Ptr_Buffer **w_inqueue;
	ff::SWSR_Ptr_Buffer *c_inqueue;
	//output queue
	ff::SWSR_Ptr_Buffer *e_outqueue;
	ff::SWSR_Ptr_Buffer *c_outqueue;
	//repository for configuration
	//void *repo;
	//for spawning new threads we will need also
	int pane_size;
	ff::ff_allocator *ffalloc; 
	int *affinities;
	int max_workers;

}controller_data_t;

/**
	Data structure passed by the various entities to the controller.
	They will contain the monitoring data referring to the last monitoring 
	step

	NOTES:
		- times are taken in cycle: for this reason frequency must be kept at maximum
*/
typedef enum monitoring_tag{
	MON_TAG=0,
	RECONF_FINISHED=1
}monitoring_tag;

typedef struct{
	union{	
		struct
		{
			char mon_id; //just for testing
			monitoring_tag tag;
			int64_t elements; //number of elements received in the last monitoring step
			int* elements_per_class; //numb of elements per class
			bool stop; //used for stopping monitoring
			char *scheduling_table; //actual scheduling table (PUNTATORE O COPIA?)
		};
		char padding[CACHE_LINE_SIZE];
	};
}emitter_monitoring_t;

typedef struct{
	union{	
		struct
		{

			//int *elements_per_class; //numb of elements per class (it is redundant but we have to keep it for the worker)
			int *computations_per_class; //numb of computation per class  (i.e. the number of pane closed)
			double *tcalcg_per_class; //tcalc expressed in usec (total tcalc per class, not on average). it refers to function G
		};
		char padding[CACHE_LINE_SIZE];
	};
}worker_monitoring_t;

typedef struct{
	union{	
		struct
		{
			monitoring_tag tag;
			char mon_id; //just for testing
			int pane_results; //total received  pane results 
			int *pane_results_per_class; //pane results received for each class (and hence it is the number of computation per class that collector will perform)
			double *tcalch_per_class; //tcalc expressed in usec (total tcalc per class, not on average). It will refers to function H
			double *latency_per_class; //expressed in usec (averaged for each class)
		};
		char padding[CACHE_LINE_SIZE];
	};
}collector_monitoring_t;


/**
	Definition of return types and grains (i.e. internal iteration) for G and H functions. Since their implementation is strictly related to
	the data structure used to keep the elements in window, their definition is in the various data structure header files
	(such as cbwindow.hpp).
*/
typedef double GReturn_type;

typedef double HReturn_type;

//result sent from the Worker towards the collector
typedef struct{
	GReturn_type res;
	int64_t id; //id of the task tha has triggered the computation
	bool isEOS; //true if it will represent to the collector the end of the stream
	int type; //class
	long long timestamp;
}WorkerResult_t;

//Function Declarations
void *emitter(void *args);
void * worker(void *args);
void * collector(void *args);
void *controller (void *args);
#endif