/**
	Version for count based windows, parallelized with the
	pane schema


	Possible Macro definitions:
	- NON_BLOCKING_FF: the internal queues are non blocking: if they are full, after some retries the program exits
	- LESS_ASSIGNED: schedules tasks according to the number of classes assigned to each worker
	- LESS_LOADED: schedules tasks according to the workload assigned to each worker
	- EMPTIEST_QUEUE: schedules tasks according to the queue length of each worker

	- TASK_BUFF: if defined the emitter has a pool of tasks used as receiving buffer (BUGGED- do not use it)
				- otherwise for each received task, the needed space is allocated by the emitter an freed by the Worker.
				  This implementation uses the Fast Flow allocator in order to reuse allocated memory
	- DONT_USE_FALLOC: if TASK_BUFF is not defined, uses standar malloc/free instead of FF allocator
	- QSPLITTED: QUEUE_SIZE will be intended as the total sum of the various queue's (Emitter-Worker). Hence the actual
				size will be QUEUE_SIZE/nworkers.
				Otherwise any queue will be initialized with QUEUE_SIZE capacity.
*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <ff/buffer.hpp>
#include <iostream>
#include <unordered_map>
#include <vector>
#include <sched.h>
#include <ff/allocator.hpp>	
#include "../Includes/RandomGenerator.h"
#include "../Includes/cycle.h"
#include "../Includes/iaperf.h"
#include "../Includes/general.h"
#include "../Includes/pane_cb.h"
#include "../Includes/affinities.h"
#include "../Includes/cbpane.hpp"
#if defined(USE_ZMQ)
#include <zmq.h>
#endif
using namespace ff;
using namespace std;

#define DEBUG(X) X;
 


//computation times for all the classes
double *comp_time;


//definition for computation time monitoring of function G: for each class, we kept the time needed for the computations
//and the number of computations (i.e. the number of pane) performed by the worker
#ifdef MONITOR_COMP_TIME
double **gcomputation_times;
int64_t **gcomputation_number;
#endif


//suffix for output files
char suffix[80];
//start time, derived from the first tuple
ticks start_global_ticks=0;

/* Standard C Function: Greatest Common Divisor */
int gcd ( int a, int b )
{
  int c;
  while ( a != 0 ) {
     c = a; a = b%a;  b = c;
  }
  return b;
}



int main(int argc, char *argv[])
{
	
	PERFINITMHZ(FREQ);

	int num_workers;
	int num_classes;
	int port;
	int i;
	int window_size;
	int window_slide;
	int pane_size;
	int input_rate=0, input_length=0;
	cpu_set_t cpuset;

	SWSR_Ptr_Buffer **quEW; //queues emitter->worker
	SWSR_Ptr_Buffer **quWC; //queues worker->collector
	//arguments for threads
	worker_data_t *worker_data;
	emitter_data_t emitter_data;
	collector_data_t collector_data;
	controller_data_t controller_data;
	//threads ids
	pthread_t *wtids;
	pthread_t ctid,cntid;
	/**
		check line arguments...
	*/
	if(argc<6)
	{
		fprintf(stderr, "Usage: %s num_classes num_workers port window_size window_slide [input_rate] [input_length]\n",argv[0] );
		return EXIT_FAILURE;
	}
	num_classes=atoi(argv[1]);
	num_workers=atoi(argv[2]);
	assert(num_workers<=max_workers);
	port=atoi(argv[3]);
	window_size=atoi(argv[4]);
	window_slide=atoi(argv[5]);
	pane_size=gcd(window_size,window_slide); //ATT: non so bene cosa cambia (correttezza, performance) se modifichiamo questo
	//pane_size=250;
	printf("Pane size: %d\n",pane_size);
	#if defined(TASK_BUFF)
		printf("Using task buffer on the emitter\n");
	#else
		#if !defined(DONT_USE_FFALLOC)
			printf("Using fast flow allocator\n");
		#else
			printf("Using standard malloc\n");
		#endif
	#endif
	//the last two are used only for testing purposes
	if(argc>=7)
		input_rate=(int)atof(argv[6]);
	if(argc>=8)
		input_length=(int)atof(argv[7]);

	/**
		Initialize data structures
	*/
	#if defined(MONITORING)
		SWSR_Ptr_Buffer *quECN; //queue emitter->controller
		SWSR_Ptr_Buffer *quCCN; //queue collector->controller
		SWSR_Ptr_Buffer **quWCN; //queues workers->controller
		SWSR_Ptr_Buffer *quCNE; //queue controller->emitter
		SWSR_Ptr_Buffer *quCNC; //queue controller->coolector
		//reconf_repository_t *repo; //repository that will contain all the structures needed for reconfigurations
	#endif
	pthread_barrier_t barrier;
	pthread_barrier_init (&barrier, NULL, num_workers + 2);
	//queues
	quEW = (SWSR_Ptr_Buffer**) malloc(max_workers * sizeof(SWSR_Ptr_Buffer*));
	ERRNULL(quEW);
	quWC = (SWSR_Ptr_Buffer**) malloc(max_workers * sizeof(SWSR_Ptr_Buffer*));
	ERRNULL(quWC);
	#ifdef QSPLITTED
	int qsize=QUEUE_SIZE/max_workers;
	#else
	int qsize=QUEUE_SIZE;
	#endif
	ticks *start_global_ticks=(ticks*)calloc(1,sizeof(ticks));

	#if defined(MONITORING)
		quWCN= (SWSR_Ptr_Buffer**) malloc(max_workers * sizeof(SWSR_Ptr_Buffer*));
		ERRNULL(quWCN);
		quECN=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
		quECN->init();
		quCCN=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
		quCCN->init();
		quCNE=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
		quCNE->init();
		quCNC=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
		quCNC->init();
	#endif

	printf("Queue size: %d\n",qsize);
	for (i = 0; i < num_workers; i++) {
		quEW[i] = new SWSR_Ptr_Buffer(qsize);
		quEW[i]->init();
		quWC[i] = new SWSR_Ptr_Buffer(qsize);
		quWC[i]->init();
		#if defined(MONITORING)
			quWCN[i]=new SWSR_Ptr_Buffer(QUEUE_SIZE_MON);
			quWCN[i]->init();
		#endif
	}
	

	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
        ff_allocator* ffalloc=new ff_allocator;
        assert(ffalloc);
        //qui indicano il numero di segmenti per dimensione 32,64,128,...
        int nslabs[N_SLABBUFFER]={0,QUEUE_SIZE*num_workers,0,0,0,0,0,0,0 };
        if (ffalloc->init(nslabs)<0) {
            error("FATAL ERROR: allocator init failed\n");
            abort();
        }
	#endif            
	//qui ci vanno altre cose....

	//suffix for output files
	sprintf(suffix,"%d_%d_%d_%d_%d_%d",num_classes,num_workers,window_size,window_slide,input_rate,input_length);
	
	/**
		Threads creations
	*/
	//Workers creation
	wtids = (pthread_t*) malloc(num_workers * sizeof(pthread_t));
	worker_data =(worker_data_t*) malloc(num_workers * sizeof(worker_data_t));
	for (i = 0; i < num_workers; i++) {		
		pthread_t tid;
		worker_data[i].workerId = i;
		worker_data[i].inqueue = quEW[i];
		worker_data[i].outqueue = quWC[i];
		worker_data[i].barrier=&barrier;
		worker_data[i].pane_size=pane_size;
		worker_data[i].freq=FREQ;
		worker_data[i].num_classes=num_classes;
		#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
			worker_data[i].ffalloc=ffalloc;
		#endif
		#if defined(MONITORING)
			worker_data[i].cn_outqueue=quWCN[i];
			//printf("QUeue pointer for worker: %d is %p\n",i,quWCN[i]);
			//worker_data[i].repo=repo;
		#endif
		pthread_create(&tid, NULL, worker, &(worker_data[i]));
		//set CPU affinity of worker threads
		CPU_ZERO(&cpuset);
		CPU_SET(affinities[i], &cpuset);

		if (pthread_setaffinity_np(tid, sizeof(cpu_set_t), &cpuset)) {
			cerr << "Cannot set thread to CPU " << i << endl;
		}
		//worker_data[i].num_workers = num_workers;
		wtids[i] = tid;
	}

	//collector creation
	collector_data.num_workers=num_workers;
	collector_data.num_classes=num_classes;
	collector_data.inqueue=quWC;
	collector_data.wqueue=quEW;
	collector_data.barrier=&barrier;
	collector_data.pane_size=pane_size;
	collector_data.window_size=window_size;
	collector_data.freq=FREQ;
	collector_data.start_global_ticks=start_global_ticks;
	collector_data.suffix=suffix;
	collector_data.window_slide=window_slide;
	#if defined(MONITORING)
		collector_data.cn_outqueue=quCCN;
		collector_data.cn_inqueue=quCNC;
	#endif
	pthread_create(&ctid, NULL, collector, &collector_data);
	//set CPU affinity of collector thread
	CPU_ZERO(&cpuset);
	CPU_SET(collector_affinity, &cpuset);

	if (pthread_setaffinity_np(ctid, sizeof(cpu_set_t), &cpuset)) {
		cerr << "Cannot set thread to CPU " << i << endl;
	}
	#if defined(MONITORING)
	//controller creation
	controller_data.num_workers=num_workers;
	controller_data.num_classes=num_classes;
	controller_data.e_inqueue=quECN;
	controller_data.w_inqueue=quWCN;
	controller_data.c_inqueue=quCCN;
	controller_data.c_outqueue=quCNC;
	controller_data.e_outqueue=quCNE;
	controller_data.suffix=suffix;
	controller_data.pane_size=pane_size;
	controller_data.affinities=affinities;
	controller_data.max_workers=max_workers;
	controller_data.freq=FREQ;
	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
		controller_data.ffalloc=ffalloc;
	#endif
	//controller_data.repo=repo;

	pthread_create(&cntid, NULL, controller, &controller_data);
	//set CPU affinity of collector thread
	CPU_ZERO(&cpuset);
	CPU_SET(controller_affinity, &cpuset);
	if (pthread_setaffinity_np(cntid, sizeof(cpu_set_t), &cpuset)) {
		cerr << "Cannot set thread to CPU " << i << endl;
	}

	#endif

	// Start working as emitter
	emitter_data.num_workers=num_workers;
	emitter_data.num_classes=num_classes;
	emitter_data.port=port;
	emitter_data.outqueue=quEW;
	emitter_data.barrier=&barrier;
	emitter_data.start_global_ticks=start_global_ticks;
	emitter_data.freq=FREQ;
	emitter_data.pane_size=pane_size;
	#if !defined(DONT_USE_FFALLOC) && !defined(TASK_BUFF)
		emitter_data.ffalloc=ffalloc;
	#endif
	#if defined(MONITORING)
		emitter_data.cn_outqueue=quECN;
		emitter_data.cn_inqueue=quCNE;
		//emitter_data.repo=repo;
	#endif
	
	// Set my affinity
	CPU_ZERO(&cpuset);
	CPU_SET(emitter_affinity, &cpuset);
	pthread_t etid = pthread_self();

	if (pthread_setaffinity_np(etid, sizeof(cpu_set_t), &cpuset)) {
		cerr << "Cannot set thread to CPU " << i << endl;
	}
	

	emitter(&emitter_data);

	// Wait for the completion of the threads
	void *retval ;
	for (i = 0; i < num_workers; i++) {
		pthread_join(wtids[i], &retval);
	}
	pthread_join(ctid, &retval);
	#if defined(MONITORING)
	pthread_join(cntid, &retval);
	#endif
	
	return EXIT_SUCCESS;
}
